import { LoggingService } from '@/services/LoggingService'
import { Application, Container, Sprite, Texture } from 'pixi.js'
import { Scale, type ScaleFunction } from './utilities/scale'
import { Tile } from './Tile'
import type { Movement, PlayerDto, TileDto } from '@/common/interfaces'
import { createContainer } from '@/common/helpers'
import { createText, playerNameText } from './utilities/fonts'

export class OtherHand {
  tilesInitialNumber: number = 7
  player?: PlayerDto
  hand: Tile[] = []
  container: Container = new Container()
  height: number
  width: number
  scale: number = 0.5
  scaleY!: ScaleFunction
  scaleX!: ScaleFunction
  x: number = 0
  y: number = 0
  grain: number = 25
  logger: LoggingService = new LoggingService()
  tilesLayer!: Container
  interactionsLayer!: Container

  constructor(
    private app: Application,
    public position: 'left' | 'right' | 'top' = 'left'
  ) {
    this.height = 100
    this.width = 300
    app.stage.addChild(this.container)
    const { x, y } = this.getPosition()
    this.container.width = this.width
    this.container.height = this.height
    this.container.x = x
    this.container.y = y
    this.calculateScale()
    this.initLayers()
  }

  setPlayer(player: PlayerDto) {
    this.player = player
    this.container.addChild(createText(`${player.name}`, this.width / 2, 12, playerNameText))
  }

  setHand(tiles: TileDto[]) {
    this.hand = tiles.map((tile) => new Tile(tile.id, this.app.ticker, undefined, this.scale))
    this.render()
  }

  update(move: Movement) {
    this.hand = this.hand.filter((tile) => tile.id !== move?.tile?.id)
    this.render()
  }

  private render() {
    this.tilesLayer.removeChildren()
    const x = -9
    this.hand.forEach((tile, index) => {
      tile.setPosition(this.scaleX(x + index * 2), this.height / 2)
      this.tilesLayer.addChild(tile.getSprite())
    })
  }

  private addBg() {
    const bg = new Sprite(Texture.WHITE)
    bg.alpha = 0.08
    bg.width = this.width
    bg.height = this.height
    this.container.addChild(bg)
  }

  private getPosition() {
    let x = 0
    let y = 0

    if (this.position === 'left') {
      x = 0
      y = 30
    } else if (this.position === 'right') {
      x = this.app.canvas.width - this.width
      y = 30
    } else {
      x = (this.app.canvas.width - this.width) / 2
      y = 0
    }
    return { x, y }
  }

  private initLayers() {
    this.container.removeChildren()
    this.addBg()
    this.tilesLayer = createContainer({
      width: this.width,
      height: this.height,
      x: 0,
      y: 0,
      parent: this.container
    })
    this.interactionsLayer = createContainer({
      width: this.width,
      height: this.height,
      x: 0,
      y: 0,
      parent: this.container
    })
    this.container.addChild(this.tilesLayer)
    this.container.addChild(this.interactionsLayer)
  }

  private calculateScale() {
    const scaleXSteps = Math.floor(this.width / (this.grain * this.scale)) / 2
    const scaleYSteps = Math.floor(this.height / (this.grain * this.scale)) / 2
    this.scaleX = Scale([-scaleXSteps, scaleXSteps], [0, this.width])
    this.scaleY = Scale([-scaleYSteps, scaleYSteps], [0, this.height])
  }
}
