import { ServiceBase } from '@/services/ServiceBase'
import { useAuthStore } from '@/stores/auth'
import { storeToRefs } from 'pinia'
import { NetworkService } from '@/services/NetworkService'
import dayjs from 'dayjs'

export class AuthenticationService extends ServiceBase {
  private networkService = new NetworkService()

  isAuthenticated() {
    const auth = useAuthStore()
    const { isLoggedIn } = storeToRefs(auth)
    return isLoggedIn
  }

  async login(username: string, password: string) {
    try {
      const res = await this.networkService.post({
        uri: '/login',
        body: { username, password }
      })
      const { token } = res
      this.persist(token)
      return token
    } catch (error) {
      console.error(error)
    }
  }

  async logout() {
    this.removePersistence()
  }

  private removePersistence() {
    const auth = useAuthStore()
    const { setJwt, setUser } = auth
    setJwt(undefined)
    setUser(undefined)
    sessionStorage.removeItem('token')
  }

  private persist(jwt: string) {
    const auth = useAuthStore()
    const { setJwt, setUser } = auth
    const loggedUser = this.parseJwt(jwt)
    setJwt(jwt)
    setUser(loggedUser)
    sessionStorage.setItem('token', jwt)
  }

  private parseJwt(token: string) {
    if (!token) {
      return
    }
    const base64Url = token.split('.')[1] ?? ''
    const base64 = base64Url.replace('-', '+').replace('_', '/')
    return JSON.parse(window.atob(base64))
  }

  fromStorage() {
    const token = sessionStorage.getItem('token')
    if (token) {
      try {
        const parsed = this.parseJwt(token)
        const isAfter = dayjs().isAfter(parsed.exp * 1000)
        if (isAfter) {
          this.removePersistence()
          return
        }
        this.persist(token)
        this.logger.debug('Token loaded from storage', parsed)
      } catch (error) {
        this.logger.error(error, 'Error parsing token')
        this.removePersistence()
      }
    } else {
      this.removePersistence()
    }
  }

  hasRoles(rolesToCheck: string[]) {
    const auth = useAuthStore()
    const { roles } = storeToRefs(auth)
    return roles.value.some((role) => rolesToCheck.includes(role))
  }
}
