import { io, Socket } from 'socket.io-client'
import { SocketIoEventManager } from '@/managers/SocketIoEventManager'
import { useAuthStore } from '@/stores/auth'
import { storeToRefs } from 'pinia'
import { ServiceBase } from './ServiceBase'

export class SocketIoClientService extends ServiceBase {
  private socket!: Socket
  private isConnected = false
  private gameEventManager: SocketIoEventManager

  constructor(private url: string) {
    super()
    this.gameEventManager = new SocketIoEventManager(this)
  }

  async connect(): Promise<void> {
    const auth = useAuthStore()
    const { jwt, isLoggedIn } = storeToRefs(auth)
    return new Promise((resolve, reject) => {
      if (!isLoggedIn) {
        reject('Not logged in')
      }
      this.socket = io(this.url, {
        auth: {
          token: jwt.value
        }
      })
      this.socket.on('connect', () => {
        if (this.socket && this.socket.recovered) {
          console.log('socket recovered succesfully')
        } else {
          console.log('socket connected')
        }
        this.isConnected = true
        this.addEvents()
        resolve()
      })
    })
  }

  addEvents(): void {
    this.socket.on('disconnect', () => {
      this.isConnected = false
      console.log('Disconnected from server')
    })

    this.socket.on('reconnect', () => {
      this.isConnected = true
      console.log('Reconnected to server')
    })

    this.socket.on('reconnect_error', () => {
      this.isConnected = false
      console.log('Failed to reconnect to server')
    })

    this.socket.on('ping', () => {
      console.log('Ping received from server')
      this.socket.emit('pong') // Send pong response
    })

    // Custom events

    // this.socket.on('makeMove', async (data: any, callback: any) => {
    //   callback(await this.gameEventManager.handleCanMakeMoveEvent(data))
    // })

    // this.socket.on('chooseTile', async (data: any, callback: any) => {
    //   callback(await this.gameEventManager.handleCanSelectTileEvent())
    // })

    this.socket.on('server:game-event', (data: any) => {
      this.gameEventManager.handleGameEvent(data)
    })

    this.socket.on('server:game-event-ack', async (data: any, callback: any) => {
      await this.gameEventManager.handleGameEventAck(data, callback)
    })
  }

  sendMessage(event: string, data: any): void {
    if (this.isConnected) {
      this.socket?.emit('client:event', { event, data })
      console.log('sendMessage :>> ', event, data)
    } else {
      console.log('Not connected to server')
    }
  }

  async sendMessageWithAck(event: string, data: any): Promise<any> {
    if (this.isConnected) {
      console.log('sendMessageWithAck :>> ', event, data)
      return await this.socket?.emitWithAck('client:event-with-ack', { event, data })
    } else {
      console.log('Not connected to server')
    }
  }

  disconnect(): void {
    this.socket?.disconnect()
    this.isConnected = false
    console.log('Disconnected from server')
  }
}
